/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */


/************************************************************************
 * sock.c: implementation of Socket class
************************************************************************/
#include "threadP.h"


void
Socket::Init( thread_t tid, PDSOCKET s, bool special, int (*will_block)(void*), void* desc )
{
	IOBindable::Init( tid, special, will_block, desc );
	sock = s;

#if defined(i386_unknown_nt4_0)
	// obtain an event object to associate with this socket
	hEvent = WSACreateEvent();
	if( hEvent == WSA_INVALID_EVENT )
	{
		thr_report_error("unable to allocate Event object\n");
		// TODO - is this fatal?
	}

    hasData = false;
#endif // defined(i386_unknown_nt4_0)
}


void
Socket::Destroy( void )
{
#if defined(i386_unknown_nt4_0)
	if( hEvent != WSA_INVALID_EVENT )
	{
		WSACloseEvent( hEvent );
	}
#endif // defined(i386_unknown_nt4_0)
	IOBindable::Destroy();
}



Socket*
Socket::Find( PDSOCKET s, thread_t* tidp )
{
	ThrTableIterator iter( *thr_tab );
	thread_t tid;


	while( iter.Next( item_t_socket, &tid ) )
	{
		// check to see if this Socket represents the
		// one we're looking for
		Socket* ps = thr_tab->GetSocket( tid );
		if( ps->sock == s )
		{
			if( tidp != NULL )
			{
				*tidp = tid;
			}
			return ps;
		}
	}
	return NULL;
}



Socket*
Socket::Alloc( PDSOCKET s,
				thread_t tid,
				bool special,
				int (*will_block)(void*),
				void* desc,
				thread_t* psid )
{
	/* allocate storage for the file */
	assert(psid != NULL);
	*psid = thr_tab->Alloc( item_t_socket );

	/* construct the entry */
	Socket* ps = thr_tab->GetSocket( *psid );
	ps->Init( tid, s, special, will_block, desc );

	return ps;
}


int
Socket::Read( void* buf, unsigned bufsize, unsigned* nbytes )
{
	int ret;

	ret = recv( sock, (char*)buf, bufsize, 0 );
	if( ret != PDSOCKET_ERROR )
	{
		*nbytes = (unsigned)ret;
#if defined( i386_unknown_nt4_0)
        hasData = false;
#endif // defined(i386_unknown_nt4_0)
	}
	return (ret != PDSOCKET_ERROR) ? THR_OKAY : PDSOCKET_ERRNO;
}


int
Socket::Write( void* buf, unsigned bufsize, unsigned* nbytes )
{
	int ret;

	ret = send( sock, (const char*)buf, bufsize, 0 );
	if( ret != PDSOCKET_ERROR )
	{
		*nbytes = (unsigned)ret;
	}
	return (ret != PDSOCKET_ERROR) ? THR_OKAY : PDSOCKET_ERRNO;
}


#if defined(THREADLIB_DEBUG)
void
Socket::WriteTo( ostream& s ) const
{
	// dump our base class data
	IOBindable::WriteTo( s );

	// dump our data
	s << "socket = " << s << endl;
}
#endif // defined(THREADLIB_DEBUG)


PDDESC
Socket::GetWaiterDesc( void ) const
{
#if defined(i386_unknown_nt4_0)
	return hEvent;
#else // defined(i386_unknown_nt4_0)
	return sock;
#endif // defined(i386_unknown_nt4_0)
}

#if defined(i386_unknown_nt4_0)

void
Socket::SetEvents( void )
{
	assert( hEvent != WSA_INVALID_EVENT );

	// set our event to signal on events of interest
	// (note that, as a side effect, this call sets
	// the socket to non-blocking mode)
	if( WSAEventSelect( sock, hEvent, FD_CONNECT | FD_ACCEPT | FD_READ | FD_CLOSE ) == SOCKET_ERROR )
    {
		thr_report_error("Failed to set socket event mask [%u]\n",
			WSAGetLastError() );
		// TODO - is this fatal?
	}
}

void
Socket::ResetEvents( void )
{
	assert( hEvent != WSA_INVALID_EVENT );

	// stop watching events
	if( WSAEventSelect( sock, hEvent, 0 ) != SOCKET_ERROR )
	{
		// reset socket to blocking mode
		u_long block = 0;
		if( ioctlsocket( sock, FIONBIO, &block ) == SOCKET_ERROR )
		{
			thr_report_error("Failed to reset socket to blocking mode [%u]\n",
				WSAGetLastError() );
			// TODO - is this fatal?
		}
	}
	else
	{
		thr_report_error("Failed to reset socket event mask [%u]\n",
			WSAGetLastError() );
		// TODO - is this fatal?
	}
}


//
// clear_ready_sock
//
// Given the socket that was used in a read or accept operation,
// update's our state associated with the Socket to indicate that
// data is no longer known to be available on that socket.
//
// Note: this ugliness is needed to assure correct interaction between
// sockets bound with the thread library as 'special' and the RPC
// package supported by Paradyn under Windows.
//
void
clear_ready_sock( PDSOCKET sock )
{
    ThrTableIterator iter( *thr_tab );
    thread_t stid;

    while( iter.Next( item_t_socket, &stid ) )
    {
        if( thr_socket( stid ) == sock )
        {
            Socket* ps = thr_tab->GetSocket( stid );
            ps->HasData( false );
            break;
        }
    }
}

#endif // defined(i386_unknown_nt4_0)


